"""
Zyncer

A Zope product that provides a synchronization service
to copy (and compare) objects from one Zope server to another.

Home page:    http://zsyncer.sourceforge.net
Release version: see version.txt
Instructions:    see README.txt
Copyright:       see LICENSE.txt
Authors:         see credits.txt
Changelog:       see changes.txt
Future Plans:    see TODO.txt

"""

# stdlib imports
import base64, sys, os, time
import traceback
from StringIO import StringIO

# zope imports
import OFS.SimpleItem
import Acquisition
import zLOG
import AccessControl.Role
import OFS.PropertyManager
from AccessControl import getSecurityManager
from AccessControl import ClassSecurityInfo, getSecurityManager, Unauthorized
from AccessControl import Unauthorized
from Acquisition import aq_parent, aq_inner, aq_base
from DateTime import DateTime
from Globals import DTMLFile, MessageDialog, Persistent, data_dir
from Globals import InitializeClass

# imports for rpc using xmlrpc
import xmlrpclibBasicAuth

# imports for rpc using ZPublisher.Client
from cPickle import loads, dumps
from ZPublisher import Client
from ZODB.POSException import ConflictError

# imports for diff
from OFS.History import html_diff, replace
from string import join, split, atoi, strip
from DocumentTemplate.DT_Util import html_quote
from DocumentTemplate.DT_Var import url_quote_plus
import types
from Config import diffable, diff_methods, descriptions, error_messages, \
                   syncable, reserved_names, colours, color_200, color_error

# imports for CMF if available
try:
    from Products.CMFDefault.DublinCore import DefaultDublinCoreImpl
except ImportError:
    DefaultDublinCoreImpl = None

# constants
OOD = 'out of date'
OK = 'ok'
EXTRA = 'extra'
MISSING = 'missing'
ZSYNC_PERMISSION = "ZSyncer: Use ZSyncer"
MGT_SCR_PERMISSION = 'View management screens'




manage_addZSyncerForm = DTMLFile('dtml/Add', globals()) 

def manage_addZSyncer(self, id='', title='', REQUEST=None):
    """adds a server
    """
    self._setObject(id, ZSyncer(id, title))
    if REQUEST is not None:
        return MessageDialog(title = 'Added', 
                             message = "ZSyncer sucessfully added.",
                             action = 'manage_main', )


class ZSyncerConfigError(Exception):
    """There is a problem with the configuration of this  ZSyncer.
    """
    
class ZSyncer(OFS.SimpleItem.Item, Persistent, Acquisition.Implicit, 
              AccessControl.Role.RoleManager,
              OFS.PropertyManager.PropertyManager):
    """A service provider that allows you to 'push' objects
    from one running Zope to another across the network.
    You need one instance on both the source and destination servers.
    """
    
    meta_type = 'ZSyncer'
    security = ClassSecurityInfo()
    manage_options = (
        {'label': 'Edit Server', 'action': 'manage_propertiesForm' },
        {'label': 'Sync', 'action': 'manage_sync' },
        {'label': 'Security', 'action': 'manage_access' },
    ) 

    _properties=(
        {'id':'title', 'type':'string', 'mode':'w'},
        {'id':'dest_servers', 'type':'lines', 'mode':'w'},
        {'id':'log', 'type':'boolean', 'mode':'w'},
        {'id':'logfile', 'type':'string', 'mode':'w'},
        {'id':'approval', 'type':'boolean', 'mode':'w'},
        {'id':'override', 'type':'boolean', 'mode':'w'},
        {'id':'user', 'type':'string', 'mode':'w'},
        {'id':'pwd', 'type':'string', 'mode':'w'},
        {'id':'syncable', 'type':'multiple selection', 'mode':'w',
         'select_variable': 'syncable'},
        {'id':'use_relative_paths', 'type':'boolean', 'mode':'w'},
        {'id':'add_syncable', 'type':'lines', 'mode':'w'},
        {'id':'connection_type', 'type':'string', 'mode':'w'},
        {'id':'filterObjects', 'type':'boolean', 'mode': 'w'},
        {'id':'filterOutObjects', 'type':'boolean', 'mode': 'w'},
        {'id':'non_syncable', 'type':'lines', 'mode':'w'},
        
        )

    # some immutable class-level defaults
    # ... these help with upgrading old instances.
    filterObjects = 0
    filterOutObjects = 0
    add_syncable = ()
    non_syncable = ()
    use_relative_paths = 0
    connection_type = 'ZPublisher.Client'
    
    security.declareProtected(ZSYNC_PERMISSION,
                              'manage_addProperty',
                              'manage_editProperties',
                              'manage_delProperties',
                              'manage_changeProperties'
                              )

    security.declareProtected(MGT_SCR_PERMISSION,
                              'manage_tabs',
                              'manage_propertiesForm')
    manage_propertiesForm = DTMLFile('dtml/Edit', globals())
    
    security.declareProtected(MGT_SCR_PERMISSION, 'manage_sync')
    manage_sync = DTMLFile('dtml/Sync', globals())

    security.declareProtected(MGT_SCR_PERMISSION, 'manage_approval')
    manage_approval = DTMLFile('dtml/Approval', globals())

    security.declareProtected(MGT_SCR_PERMISSION, 'manage_diff')
    manage_diff = DTMLFile('dtml/Diff', globals())

    security.declareProtected(MGT_SCR_PERMISSION, 'manage_folders')
    manage_folders = DTMLFile('dtml/Folder', globals())


    # immutable default attributes for upgrades
    connection_type = 'ZPublisher.Client'
    syncable = tuple(syncable)
    dest_servers = ()
    
    def __init__(self, id, title=''):
        """initialize a new instance of Server
        """
        self.id = id
        self.title = title
        self.source_server = [ None, ] # source is this Syncer. XXX unused?
        self.dest_servers = [ ]
        self.log = 0
        self.approval = 0
        self.override = 0
        self.user = ''  
        self.pwd = ''   
        self.syncable = syncable
        self.logfile = os.path.join(data_dir, 'ZSyncer.log')
        self.connection_type = 'ZPublisher.Client'
        self.use_relative_paths = 1
        self.filterObjects = 0
        self.filterOutObjects = 0
        self.add_syncable=[]
        self.non_syncable = []

    # python2.2 adds dosctrings to strings. This means
    # that if you use zope <= 2.6.? and python >= 2.2,
    # these sensitive attributes are directly publishable!
    # Odd combination, but protect them to be safe.
    security.declarePrivate('user')
    security.declarePrivate('pwd')

    security.declareProtected(MGT_SCR_PERMISSION, 'manage_approvedAction') 
    def manage_approvedAction(self, REQUEST=None, action='',
                              object_ids=None, syncer=None, approver=None,
                              comments=None):
        """If this action is to be approved, throw that data in
        into the message list
        """
        msgs = ['User: %s' % syncer,]
        if self.approval:
            msgs.append('Approved by: %s' % approver)

        # strip out any newlines in comments
        comments = comments.replace('\n', '')
        msgs.append('Comments: %s' % comments)

        # call the relevant thing
        if action == 'manage_pushToRemote': 
            return self.manage_pushToRemote(object_ids, msgs=msgs,
                                            REQUEST=REQUEST)
        elif action == 'manage_deleteRemote': 
            return self.manage_deleteRemote(object_ids, msgs=msgs,
                                            REQUEST=REQUEST)
        else: 
            return self._error(msg='Unknown action')

        
    
    security.declareProtected(ZSYNC_PERMISSION, 'manage_deleteRemote')
    def manage_deleteRemote(self, object_ids, msgs=None, REQUEST=None):
        """Interface to calling delete on each destination server.
        """
        if msgs is None:
            msgs = []
        if type(object_ids) == types.StringType:
            object_ids = [object_ids,]
        for object_id in object_ids:
            for srv in self.dest_servers:
                # more logging
                msgs.append('Object %s deleting from %s' % (object_id, srv))
                msgs.append(self._deleteRemote(srv, object_id))
        self._do_messages(msgs, REQUEST)

    security.declareProtected(ZSYNC_PERMISSION, 'manage_pushToRemote')
    def manage_pushToRemote(self, object_ids, msgs=None, REQUEST=None):
        """Interface to pushing each object to destination
        server.
        """
        if type(object_ids) == types.StringType:
            object_ids = [object_ids,]
            
        # guestimate the size of output status messages so
        # we can stream it w/ response.write().
        # XXX is it possible w/ javascript to force the browser to scroll down?
        content_length=1000 + (200*len(self.dest_servers) * len(object_ids))
        # be sure to log something even if syncing fails.
        if REQUEST and self.log:
            user = getSecurityManager().getUser()
            self._log(self._get_time(),
                      '%s attempting to sync %s\n' % (user, object_ids))
        written = self._msg_header(REQUEST=REQUEST,
                                   content_length=content_length)
        # handle user & approver comments
        if msgs is None:
            msgs = []  
        for m in msgs:
            written += self._do_one_msg(TextMsg(m), REQUEST)
        # sync
        for object_id in object_ids:
            for srv in self.dest_servers:
                # more logging
                status = self._exportRemote(srv, object_id)
                msg = StatusMsg('Object %s to %s' % (object_id, srv),
                                status)
                written += self._do_one_msg(msg, REQUEST)
        self._msg_footer(REQUEST, content_length - written)


    security.declareProtected(ZSYNC_PERMISSION, 'manage_diffObject')
    def manage_diffObject(self, object_id, REQUEST=None):
        """Try to get the object from the destination server
        and diff on it.
        Return the result as a dictionary.
        """
        error = ''
        source = self.manage_getSource(object_id)
        dest_server = self._getFirstDestination()
        try:
            dest = self._srcRemote(dest_server, object_id)
        except:
            # XXX except what???
            # XXX FIXME is there anything we can do about this? It kinda sucks
            # that non-ascii chars prevent diffing.
            # Was this an artifact of XML-RPC? Test again with ZPub.Client.
            error = "Couldn't get source of remote object - "
            error +="it might contain non-ASCII values?" 
            dest = ''
        # check result, there should be a better way to do this using raise, 
        # oh well.
        if type(source) == types.IntType:
            error = StatusMsg("Getting local source:", source).html()
        if type(dest) == types.IntType:
            error = StatusMsg("Getting remote source:", dest).html()
        if error:
            diff = ''
        else:
            diff = html_diff(dest, source)
        return {'diff': diff, 'source': source, 'dest': dest, 'error': error}

    security.declareProtected(MGT_SCR_PERMISSION, 'manage_compare')
    def manage_compare(self, path, recurse=None, REQUEST=None):
        '''Compare items, optionally recursing into subfolders;
        optionally filtering by status according to REQUEST.
        Return a list of dictionaries suitable
        for viewing on the source server.  Sort by path.
        '''
        default_filter = (MISSING, OK, OOD, EXTRA)
        if REQUEST: 
            status_filter = REQUEST.get('show_status') or default_filter
        results = self._manage_compare_path(path, recurse=recurse,
                                            REQUEST=REQUEST)
        # filter once after getting all results.
        if status_filter != default_filter:
            filtered = [r for r in results if r['status'] in status_filter]
            results = filtered
        # sort by path.
        sortable = [('/'.join(r['path']), r) for r in results]
        sortable.sort()
        results = [s[1] for s in sortable]
        return results

    security.declareProtected(MGT_SCR_PERMISSION, 'manage_getSource')
    def manage_getSource(self, obj_path):
        """Get a src from an object (if allowed) suitable for diffing.
        """
        try:
            obj = self._getObject(obj_path)
        except KeyError:
            return 404

        c = getSecurityManager().checkPermission
        if not c(MGT_SCR_PERMISSION, obj):
            return 403
        # try to handle various kinds of objects...
        # use whatever's in the config for this meta_type.
        base_obj = aq_base(obj)
        mt = base_obj.meta_type
        attrname = diff_methods.get(mt, diff_methods['DEFAULT'])
        diff_attr = getattr(base_obj, attrname)
        if callable(diff_attr):
            src = diff_attr()
        else:
            src = diff_attr
        return src

    
    security.declareProtected(ZSYNC_PERMISSION, 'manage_replaceObject')
    def manage_replaceObject(self, obj_path, data=None, decode=0):
        """
        If an object already exists at *obj_path*,  delete it.
        If *data* is not None, treat is as a picked new object to
        add at that path.
        """
        # make sure we always get a string or a sequence 
        if type(obj_path) == types.StringType:
            obj_url = obj_path
            obj_path = obj_path.split('/')
        else: 
            obj_path = list(obj_path)
            obj_url = '/'.join(obj_path)
        # let's try finding the parent
        container = aq_parent(aq_inner(self))
        try:
            parent = container.restrictedTraverse(obj_path[:-1])
        except KeyError:
            return 404

        # let's check they are allowed to do this
        c = getSecurityManager().checkPermission
        allowed = 1
        for perm in ['Delete objects', 'Import/Export objects']:
            if not c(perm, parent):
                allowed = 0
        if not allowed:
            return 403

        # If the parent supports ordering, note the old object's position 
        # so that we can move the new object to that position.
        position = None
        # Check for Zope 2.7+ OrderSupport
        has_order_support = getattr(parent.aq_explicit, 'has_order_support', 0)
        # Check for OrderedObjectManager monkey patch
        has_ordered_object_manager = getattr(parent.aq_explicit, 'move_objects_by_positions', None)

        # if there is one there already, delete it.
        # hooray for transactions :)
        if obj_path[-1] in parent.objectIds(): 
            if has_order_support:
                position = parent.getObjectPosition(obj_path[-1])
            elif has_ordered_object_manager:
                position = parent.get_object_position(obj_path[-1])

            parent.manage_delObjects([obj_path[-1],])
        else:
            if data is None: # ok if we are deleting
                return 404

        # Add the new object, if any.
        if data is not None:
            zLOG.LOG('ZSyncer', -100, ' trying to import %s **' % obj_url)
            # fake a file using StringIO
            # XXX this is very memory-intensive if we get a lot of data.
            file = StringIO()
            # did we get encoded data?
            if decode:
                file.write(self._decode(data))
                zLOG.LOG('ZSyncer', -100, 'decoding, as instructed')
            else:
                file.write(data)
                zLOG.LOG('ZSyncer', -100, 'NOT decoding, as instructed')
            file.seek(0)

            # now import
            try: 
                new = parent._p_jar.importFile(file)
            except:
                # we'll re-raise it
                msg = '_p_Jar.importFile(file) failed '
                msg += '... receiving %s' % obj_url
                zLOG.LOG('ZSyncer', 200, msg)
                raise
            object_id = new.getId()
            try: 
                parent._setObject(object_id, new)
            except:
                # we'll re-raise it
                msg = 'parent._setObject failed '
                msg += '...receiving %s' % obj_url                
                zLOG.LOG('ZSyncer', 200, msg)
                raise

            if(position is not None):
                try: 
                    methodname = 'moveObjectToPosition'
                    if has_order_support:
                        parent.moveObjectToPosition(object_id, position)
                    elif has_ordered_object_manager:
                        methodname = 'move_objects_by_positions'
                        newposition = parent.get_object_position(object_id)
                        parent.move_objects_by_positions(object_id, position - newposition)
                except:
                    # we'll re-raise it
                    msg = 'parent.%s failed ' % methodname
                    msg += '...receiving %s' % obj_url                
                    zLOG.LOG('ZSyncer', 200, msg)
                    raise
  
        # yay!
        zLOG.LOG('ZSyncer', -100,
                 'manage_replaceObject succeeded at %s\n' % obj_url)
        return 200

    security.declareProtected(MGT_SCR_PERMISSION, 'status_colour')
    def status_colour(self, status, default="white"):
        '''gives a list of status colours for pretty html
        '''
        return colours.get(status, default)

    security.declareProtected(MGT_SCR_PERMISSION, 'status_descr')
    def status_descr(self, status):
        '''get description of each status from the config
        '''
        return descriptions.get(status, descriptions.get('default', ''))

    security.declareProtected(MGT_SCR_PERMISSION, 'manage_listObjects')
    def manage_listObjects(self, path):
        """Gets a dictionary of a folder's objects, keyed by id. 
	If path is not folderish, get a dictionary containing only 
	the path object. (so manage_compare can work on 
	non-objectManagers).
        """
        obs = {}
        # go find the object at the path
        path_object = self._getObject(path)
        path = self._getRelativePhysicalPath(path_object)
        # get the sub-objects 
        if aq_base(path_object).isPrincipiaFolderish:
            sync_objects = path_object.objectValues()
        else: 
            sync_objects = [path_object,]
        # iterate
        for ob in sync_objects:
            # we don't sync ZSyncers, that way leads to much confusion
            # so we'll just sync objects we know work.
            # avoid broken objects.
            # allow people to turn this feature on and off.
            if self.filterObjects and not ob.meta_type in self.syncable:
                continue
            if self.filterOutObjects and ob.meta_type in self.non_syncable:
                continue
            if ob.meta_type[:14] == 'Broken Because':
                continue
            if self._isReserved(ob):
                continue

            o = {}
            o['id'] = ob.getId()
            # when this is called on remote server, we can't
            # enforce that ob is in the same part of the tree as
            # self, so strict=0.  XXXX not sure if this is correct??
            o['path'] = self._getRelativePhysicalPath(ob)
            # ... was ob.getPhysicalPath()
            
            # this should always be a string right?
            o['meta_type'] = ob.meta_type

            # CMF fix, where apparently ob.icon may be callable
            o['icon'] = ob.icon
            if not isinstance(o['icon'], types.StringType):
                o['icon'] = ob.icon()
           
            o['is_folder'] = getattr(ob, 'isPrincipiaFolderish', 0)

            # Try to get DublinCore mod time, if available.
            base_ob = aq_base(ob)
            if DefaultDublinCoreImpl is not None and \
                   isinstance(base_ob, DefaultDublinCoreImpl):
                o['dc_modtime'] = base_ob.modified()
            else:
                o['dc_modtime'] = ''

            # formerly known as 'modtime'...
            bobotime = ob.bobobase_modification_time()
            o['last_modified_time'] = bobotime  

            # size in kB. For backward compatibility,
            # don't depend on this.
            try:
                o['size'] = int(ob.get_size()) / 1024
            except AttributeError:
                o['size'] = ""

            # add this to the dict of objects
            obs[o['id']] = o
        return obs

    security.declareProtected(MGT_SCR_PERMISSION, 'is_diffable')
    def is_diffable(self, meta_type=None):
        '''is this diffable
        '''
        if meta_type in diffable:
            return 1
        else:
            return 0
        
    security.declareProtected(ZSYNC_PERMISSION, 'getPath')
    def getPath(self, path):
        """get relative path to an object as a string,
        useful for building UI.
        If object doesn't exist locally, assume it's remote and
        just return the passed-in path as a string.
        """
        try:
            path = self._getRelativePhysicalPath(path)
        except KeyError:
            pass
        path = '/'.join(path)
        return path
        
##     security.declareProtected(ZSYNC_PERMISSION, 'getURLPath')
##     def getURLPath(self, path):
##         """
##         Get a virtual-host-safe absolute URL path, given a
##         (possibly relative) path to an object.

##         Note that this is not suitable for traversal
##         on the remote server, as the path to the zsyncer container
##         may be different, which matters when using relative paths.
        
##         XXX well that was its whole purpose, so, forget it :-)
##         use getPath() instead.
##         """
        
##         if type(path) == StringType:
##             path = [path,]
##         else:
##             path=list(path)
##         if self.use_relative_paths:
##             base_obj = aq_parent(aq_inner(self))
##         else:
##             base_obj = self.getPhysicalRoot()
##         basepath = base_obj.absolute_url_path()
##         if basepath == '/':
##             basepath = ''
##         path = '/'.join(path)
##         path = '%s/%s' % (base_obj.absolute_url_path(),
##                           path)
##         while path.count('//'):
##             path = path.replace('//', '/')
##         return path

    #### Diagnostic helpers #################
    

    security.declareProtected(ZSYNC_PERMISSION, 'getErrorAdvice')
    def getErrorAdvice(self, error_type, error_value):
        """Given an error_type and error_value,
        say something helpful :-)
        
        This is a bit hacky because of having two connection methods
        that do different things.
        """
        error_value = str(error_value)
        error_type = str(error_type)

        # Connection Refused
        if error_value.count('Connection refused'):
            text = ['Check that the destination server is running.',
                    'Check that routing, firewall etc. allows http ' \
                    'connections to %s' % self._getFirstDestination(),
                    ]
        else:
            text = ['The remote syncer is responding, but...']
            
        # Not Found        
        if error_value.count('NotFound') or \
               error_value.count('Not Found'):
            text += ['Check that the destination zsyncer is really there at '+ 
                    '"%s."' % self._getFirstDestination()]

        # UNAUTHORIZED
        elif error_value.count('401 Unauthorized'):
            # that works for both connection types.
            uid = self._get_user()['user']
            text+=['For the user "%s":' % uid,]
            if self.override:
                text.extend(['Check that the user exists on this server',
                             'Check the password is correct on this server'])
            text.extend(['Check that the user exists on destination server',
                         'Check the password on the destination server',
                         'Check the user has permission to use ZSyncer on' \
                         ' destination server',
                         ]
                        )

        # You're inside a missing folder
        elif error_type == 'KeyError' or \
                 error_value.count('KeyError'):
            text+=["The folder you're in may be missing on destination server",
                   'Try going up to the parent folder and sync it from there.']

        # No idea.
        else:
            text += ['Unexpected problem, maybe on the remote server.' \
                     " See below, and check the destination's error log."]
        return text
            
        
        
    ########################################################################
    # internal only
    # protected by having _ as the first character
    ########################################################################

    def _isReserved(self, ob):
        """stuff to prevent syncing in the zope root.
        """
        path = ob.getPhysicalPath()
        # don't sync yourself.
        if path == self.getPhysicalPath():
            return 1
        # names from the config.
        if len(path) == 2 and path[1] in reserved_names:
            return 1
        else:
            return 0

    def _normalize_time(self, datetime, gmt=None):
        '''convert DateTime to string representing time in local zone
        (by default), or optionally in GMT.
        Use the ls -l format (i.e. only show year if old enough),
        like "Aug  5  2003" or "Sep 21 23:22".
        '''
        if type(datetime) == types.StringType:
            return datetime  # it's already a string
        if gmt:
            datetime = DateTime(datetime, 'GMT')
        show_year = not datetime.isCurrentYear()
        if show_year:
            fmt = r'%b %d  %Y'
        else:
            fmt = r'%b %d %H:%M'
        return datetime.strftime(fmt)


    #################################################################
    # COMPARISONS
    #################################################################
    
    def _get_status(self, key, s_dict={}, d_dict={}):
        '''compare a subdict (representing an object) of two dicts
        (each representing a batch of objects), decide what the status is, 
        and create a composite dict with all info needed for display.
        Returns (status as a string, itemdict) 
        '''
        s_item, d_item = s_dict.get(key), d_dict.get(key)
        item = {}
        # Fix up types.
        if not s_item:
            s_item = {}
        if not d_item:
            d_item = {}
        # Get mod times.
        # bobobase_modification_times for last resort
        s_bobotime = s_item.get('last_modified_time', '')
        d_bobotime = d_item.get('last_modified_time', '')
        # see if we have DublinCore metadata times available
        s_dc_modtime = s_item.get('dc_modtime', '')
        d_dc_modtime = d_item.get('dc_modtime', '')

        # start looking for status.
        if not s_item and not d_item:
            zLOG.LOG('Zsyncer', 100,
                     's_item and d_item both none? impossible!')
        elif not d_item:
            item = s_item.copy()
            item['status'] = MISSING

        elif not s_item:
            item = d_item.copy()
            item['status'] = EXTRA

        else:
            item = s_item.copy()

        # Decide which timestamp to use by default for display.
        item['dest_best_modtime']=self._normalize_time(
            d_dc_modtime or d_bobotime)
        item['src_best_modtime']=self._normalize_time(
            s_dc_modtime or s_bobotime)
        item['dest_mod_time'] = d_bobotime
        item['src_mod_time'] = s_bobotime

        # If status is set, we're done.
        if item.get('status'):
            return (item['status'], item)
        
        # Compare timestamps. Note that DC timestamps MUST compare equal,
        # but bobobase timestamps are only tested for <. 
        if s_dc_modtime and d_dc_modtime:
            if s_dc_modtime == d_dc_modtime:
                item['status'] = OK
            else: 
                # XXX FIXME- for DC we should really treat < > differently,
                # which would require adding another status.
                item['status'] = OOD
        # Last resort, use bobobase_modification_time.
        elif d_bobotime < s_bobotime:
            item['status'] = OOD
        else:
            item['status'] = OK
            
        return (item['status'], item)

    def _getFirstDestination(self):
        """get the first remote server, for comparisons.
        If there are none, raise ZSyncerConfigError.
        """
        try:
            remote = self.dest_servers[0]
        except IndexError:
            raise ZSyncerConfigError, "No remote servers are configured."
        return remote
    
    def _manage_compare_path(self, path, recurse=None, REQUEST=None):
        '''Get the status for everything on the path, optionally 
        recursing into folders. Result is NOT filtered by status.
        Returns a list of dictionaries, each of which
        represents a compared object.
        '''
        compared = []
        # make sure path is usable.
        container = aq_parent(aq_inner(self))
        obj = container.restrictedTraverse(path)
        path = '/'.join(self._getRelativePhysicalPath(obj))
        # get the remote stuff
        remote = self._getFirstDestination()
        dest_dict = self._getRemoteList(remote, path)
        # get the local stuff
        local_dict = self.manage_listObjects(path)
        all_keys = {}
        for k in local_dict.keys() + dest_dict.keys():
            all_keys[k] = 1
        for key in all_keys.keys():
            status, item = self._get_status(key, local_dict, dest_dict)
            compared.append(item)
            # recurse?
            if not recurse:
                continue
            # ok, recurse.
            if item['is_folder'] and status not in (MISSING, EXTRA):
                sub_path = _normalizeRelativeURL('/'.join(item['path']))
                # Never recurse back into yourself!
                if path == sub_path:
                    continue
                #ok, really recurse, but don't filter.
                try:
                    zLOG.LOG('zsyncer', -100, "recursing",
                             "FROM %s\n TO %s" % (path, sub_path))
                    results = self._manage_compare_path(path=sub_path,
                                                        recurse=recurse,
                                                        REQUEST=REQUEST)
                except Unauthorized:
                    zLOG.LOG('zsyncer', -100, "got Unauthorized at %s" % path)
                    results = []
                compared.extend(results)
        return compared


    def _error(self, code=None, msg=None):
        # oops
        if msg is None and code is not None: 
            msg = error_messages.get(code, 'Unknown error occured')
        return MessageDialog(title='Message', message=msg,
                             action='manage_main')

    def _get_user(self):
        # get user information
        if self.override:
            _u = {'user':self.user, 'pwd':self.pwd }
        else:
            u = getSecurityManager().getUser()
            _u = {'user':u.getId(), 'pwd':u._getPassword()}
        return _u


    #####################################################################
    #  MESSAGES AND LOGGING
    #####################################################################
    
    def _get_time(self):
        # get time for logging
        # could be done using DateTime, but i think I want to fiddle this
        return time.asctime(time.localtime(time.time()))


        
    def _do_messages(self, msgs, REQUEST=None):
        """log a list of messages, and if there is a
        REQUEST, do an html display
        """
        do_log = self.log or self.approval
        do_html = REQUEST is not None
        ms = []
        # header
        size_to_write = contest_length=200 + 100*len(msgs)
        written = self._msg_header(REQUEST, size_to_write)
        
        # the actual messages now...
        processed_msgs = []
        for m in msgs:
            if type(m) == types.IntType:
                # hack-o-rama, if i have a number, it's a status code, 
                # so let's combine that with the previous message to
                # get a StatusMessage.
                try:
                    processed_msgs[-1] = StatusMsg(processed_msgs[-1], m)
                except IndexError:
                    pass
                    #print "no prev. message... this shouldn't happen"
            else:
                # plain text message
                processed_msgs.append(TextMsg(m))

        # ready to do all the messages
        for m in processed_msgs:
            written += self._do_one_msg(m, REQUEST)
            
        # footer
        self._msg_footer(REQUEST, size_to_write - written)

    def _msg_header(self, REQUEST=None, content_length=1000):
        '''writes log and/or html info at beginning of a sync.
        Returns the # of characters written in the html so we
        can keep track of streaming.
        '''
        written = 0
        if REQUEST is not None:
            # be sure we have enough content-length... overestimate.
            REQUEST.RESPONSE.setHeader('content-type', 'text/html')
            REQUEST.RESPONSE.setHeader('content-length', str(content_length),
                                       )
            head ='''
            <html><body><form action="manage_sync" method="GET">
            <input type="hidden" name="folder"
             value="%s">''' % self.REQUEST.get('folder', '/')

            REQUEST.RESPONSE.write(head)
            written += len(head)
            
        if self.log or self.approval:
            self._log(self._get_time(),
                      '-' * 40 +  '\nStarted syncing.\n')
        return written
    
    def _msg_footer(self, REQUEST, padding=0):
        if REQUEST is not None:
            foot = '''<div>
                <b>DONE!</b>
                <br />
                <input type="submit" value=" OK ">
                </div>
                </form></body></html>
            '''
            REQUEST.RESPONSE.write(foot)
            padding -= len(foot)
            REQUEST.RESPONSE.write(' ' * padding)
        if self.log or self.approval:
            self._log(self._get_time(), 'Done\n')

    def _do_one_msg(self, msg, REQUEST=None):
        """log and/or display a single Msg.
        return lenght of written html, if any.
        """
        t = self._get_time()
        self._log(t, msg)
        if REQUEST is None:
            return 0
        # stream it
        html = msg.html()
        REQUEST.RESPONSE.write(html)
        return len(html)            

    def _log(self, m_time, msgs):
        """log messages for ZSyncer
        """
        # this will become more configurable
        logfile = open(self.logfile, 'a')
        logfile.write('%s\t%s\n' % (m_time, str(msgs)))
        logfile.close()

    ######################################################################
    # Control of the remote server 
    #####################################################################

    def _getServerConn(self, server_url):
        """get a connection to the server.
        OK to leave off http://
        """
        if server_url[:7] != 'http://' and server_url[:8] != 'https://':
            server_url = 'http://%s' % server_url
        if server_url[:5] != 'http:' and server_url[:6] != 'https:':
            server_url = 'http:%s' % server_url
        return self._serverConn(server_url)

    def _serverConn(self, server_url):
        """connect to the remote server using our connection_type
        """
        u = self._get_user()
        if self.connection_type == 'xml-rpc':
            server = xmlrpclibBasicAuth.Server(server_url, u['user'], u['pwd'])
        else:
            server = _Server(server=server_url, user=u['user'],
                             passwd=u['pwd'])
        return server

                         
    def _encode(self, s):
        """encoding of transferred data
        """
        # do any other encoding here
        # eg twofish, zlib etc
        # this will do me for now
        if self.connection_type == 'xml-rpc':
            encoded = base64.encodestring(s)
        else:
            encoded = s
        return encoded

    def _decode(self, s):
        """decoding of transferred data
        """
        # do any other decoding here
        return base64.decodestring(s)


    def _getRemoteList(self, server_url, path):
        '''get the list of remote stuff
        '''
        serverconn = self._getServerConn(server_url)
        return serverconn.manage_listObjects(path)

    
    def _srcRemote(self, server_url, object_id):
        """get the source text of the remote object
        """
        serverconn = self._getServerConn(server_url)
        return serverconn.manage_getSource(object_id)

    def _deleteRemote(self, server_url, object_id):
        """deletes an object on the remote server.
        """
        serverconn = self._getServerConn(server_url)
        try:
            result = serverconn.manage_replaceObject(object_id, None,
                                                     decode=0)
        except:
            # XXX except what???
            result = 500
        return result

    def _exportRemote(self, server_url, path):
        """adds an object to the remote server
        """
        # go get the local object
        obj = self._getObject(path)
        data = StringIO()
        obj._p_jar.exportFile(obj._p_oid, data)
        encoded_data = self._encode(data.getvalue())
        del(data)
        serverconn = self._getServerConn(server_url)
        # Decide whether to do base64 decoding on the remote end.
        if self.connection_type == 'xml-rpc':
            remote_decode = 1
        else:
            remote_decode = 0
        zLOG.LOG('ZSyncer', -100, 'telling remote to decode? %d'
                 % remote_decode)
        obj_path = self._getRelativePhysicalPath(obj)
        # ok, export to the destination
        try:
            result = serverconn.manage_replaceObject(obj_path, encoded_data,
                                                     remote_decode)
        except:
            # XXX except what???
            result = 500

        return result

    def _callRemote(self, server_url, path, method_name, *args, **kw):
        """Find an arbitrary callable at *path* on the remote server,
        call it with the given arguments, and return the result.
        """
        server = self._getServerConn(server_url)
        result = server.callMethod_(path, method_name, *args, **kw)
        return result

    security.declareProtected(ZSYNC_PERMISSION, 'callMethod_')
    def callMethod_(self, path, method_name, *args, **kw):
        """Call an arbitrary method (with security checks).
        *path* is path to the object, *method_name* is method
        to call. Any result is returned.

        This is intended to be called on the remote server,
        ONLY via _callRemote().

        Note that for security reasons, you must have permission
        to use ZSyncer *and* to call the requested method in
        context.
        """
        obj = self.restrictedTraverse(path)
        method = getattr(obj, method_name)
        sm = getSecurityManager()
        if not sm.validateValue(method):
            user = sm.getUser().getId()
            err ="User %s is not allowed to call %s here" % (user,
                                                             method_name)
            raise Unauthorized, err
        result = method(*args, **kw)
        return result

    ###################################################################
    # Local object helpers
    ###################################################################
    
    def _getObject(self, path):
        """
        Get an object at the given (possibly relative) path.
        """
        p = path.strip()
        if not p and not self.use_relative_paths:
            p = '/'
        container = aq_parent(aq_inner(self))
        obj = container.restrictedTraverse(p)
        return obj
        
    def _getRelativePhysicalPath(self, obj_or_path, strict=1):
        """
        Get a clean *physical* path to the object or path, either
        relative to this zsyncer's container or relative to the app root,
        depending on setting of the use_relative_paths property.
        If *obj_or_path* is a path, it can be either a string or
        a tuple.

        Return value is a sequence of path elements, e.g.
        as returned by getPhysicalPath().
        
        If 'strict' is true, object must be somewhere within the
        container or we'll raise ValueError. True by default.

        If object doesn't exist, the underlying error is propagated
        (probably KeyError).
        """
        if type(obj_or_path) in (types.TupleType, types.ListType):
            obj_or_path = '/'.join(obj_or_path)
        if type(obj_or_path) == types.StringType:
            obj = self._getObject(obj_or_path)
        else:
            obj = obj_or_path
        obj_path = obj.getPhysicalPath()
        if not self.use_relative_paths:
            return obj_path
        container = aq_parent(aq_inner(self)) 
        container_path = container.getPhysicalPath()
        cl = len(container_path)
        if strict and obj_path[:cl] != container_path:
            raise ValueError, "%s is not within %s" % ('/'.join(obj_path),
                                                       '/'.join(container_path)
                                                       )
        return obj_path[cl:]
        
        
    ########################################################################
    ## Non-XML-RPC use thanks to Dieter Maurer.
    #  XML-RPC is crap in a non-ASCII, non-fully Unicode environment.
    #   We replace XML-RPC by "ZPublisher.Client" (on call) and
    #   "cPickle" on return
    ########################################################################
    def call_(self,request):
        '''decode a pickled *request* and execute it,
        returning the result as a pickled string.
        '''
        method,args,kw= loads(request)
        m= getattr(self,method)
        if not getSecurityManager().validateValue(m):
            raise Unauthorized
        try:
            r= (1, m(*args, **kw))
        except ConflictError:
            # this will be retried
            raise
        except:
            # XXX except what??
            # we want to send back the exception
            r= (0, sys.exc_info()[:2])
        return dumps(r)
 

 
InitializeClass(ZSyncer)
 
######################################################################
# Classes for avoiding XML-RPC
######################################################################

class _Server:
    '''Auxiliary class to call the server.

    Used for http (non-xmlrpc) syncing.
    '''
    
    def __init__(self, server, user, passwd):
        '''
        server:
        user:
        passwd: plain-text passwd
        '''
        self._info = (server, user, passwd)
        
    def __getattr__(self,method):
        return _MethodProxy(self._info, method)
 
class _MethodProxy:
    '''Wraps a method so we can call it on a remote server.
    '''
    def __init__(self,info,method):
        '''
        info = (server, user, password)
        method = the method to wrap
        '''
        self._info= info
        self._method= method
 
    def __call__(self, *args, **kw):
        server, user, password= self._info
        f= Client.Function(server+'/call_',
                           method= 'POST',
                           username= user,
                           password= password,
                           )
        f.headers['Content-Type']= 'multipart/form-data'
        r= f(request= dumps((self._method,args,kw),1))
        ok,rd= loads(r[1])
        if ok:
            return rd
        raise rd[0], rd[1]
 
##########################################################################
# Logging & UI messages
##########################################################################

class TextMsg:
    """for logging & output of arbitrary text
    """
    color = 'black'
    def __init__(self, msg):
        self.msg = msg
        
    def __str__(self):
        return self.msg
   
    def html(self):
        return '<div style="color: %s">%s</div>\n' % (self.color,
                                                      html_quote(self.msg))

class StatusMsg(TextMsg):
    """for logging & output of remote call results
    """
    def __init__(self, msg, status):
        self.status = status
        if int(status) == 200:
            self.color = color_200
            self.msg = 'OK: %s' % msg
        else:
            self.color = color_error
            self.msg = '%s: %s' % (error_messages.get(str(status),
                                                      'Unknown error!'),
                                   msg)

############################################################################
# Utility classes & functions
############################################################################

def _normalizeRelativeURL(url):
    """
    utility function - does basically what os.path.normpath()
    does for filesystem paths, but for URLs instead.
    Does not verify whether anything exists at the path.
    """
    url = url.strip()
    elements = url.split('/')
    if not elements:
        return '/'
    norm_elements = []
    for e in elements:
        e = e.strip()
        if e in ('', '.'):
            continue
        elif e == '..':
            try:
                norm_elements.pop()
            except IndexError:
                continue
        else:
            norm_elements.append(e)
    result = '/'.join(norm_elements)
    # we've thrown away any leading slashes, so put one back if needed.
    if url and url[0] == '/':
        result = '/' + result
    return result


        
